// Slack API Interface

#ifndef CHAT_SLACK_INCLUDED
#define CHAT_SLACK_INCLUDED
#include <curl/curl.h>

#include "ed_array.h"
#include "string.h"

typedef struct {
    CURL *curl;

    string token;
    string cookie;
} slack_client;

typedef struct {
    string url;
    string team;
    string user;
    string team_id;
    string user_id;
} slack_user_info;

typedef struct {
    string id;
    string name;

    string topic;
    string purpose;
} slack_channel;
arrayTemplate(slack_channel);

typedef enum {
    SLACK_MESSAGE_TYPE_MESSAGE,
} slack_message_type_t;

typedef struct {
    slack_message_type_t type;
} slack_message_type;

typedef struct {
    string user;
    string text;
} slack_message_normal;

typedef struct {
    slack_message_type type;
    union {
        slack_message_normal normal;
    };
} slack_message;
arrayTemplate(slack_message);

// TODO: merge with `slack_user_info`
typedef struct {
    string id;
    string name;
    string real_name;
} slack_user;
arrayTemplate(slack_user);

slack_user_info slack_api_auth_test(slack_client *client);
array(slack_channel)
    slack_api_channel_list(slack_client *client, slack_user_info user);
array(slack_message)
    slack_api_message_list(slack_client *client, string channel_id);
array(slack_user) slack_api_user_list(slack_client *client, string team_id);

slack_client slack_init_client();

#ifdef CHAT_SLACK_IMPLEMENTATION
#include <cJSON/cJSON.h>

#include <cJSON/cJSON.c>

static size_t write_memory_callback(void *contents, size_t size, size_t nmemb,
                                    void *userp) {
    size_t real_size = size * nmemb;
    pushArrayMulti(uint8_t, userp, contents, real_size);
    return real_size;
}

slack_user_info slack_api_auth_test(slack_client *client) {
    if (client->curl) {
        struct curl_slist *headers = NULL;
        headers = curl_slist_append(
            headers, "Content-Type: application/x-www-form-urlencoded");
        headers = curl_slist_append(headers, client->cookie.data);

        curl_easy_setopt(client->curl, CURLOPT_URL,
                         "https://slack.com/api/auth.test");

        curl_easy_setopt(client->curl, CURLOPT_HTTPAUTH, (long)CURLAUTH_BEARER);
        curl_easy_setopt(client->curl, CURLOPT_XOAUTH2_BEARER,
                         client->token.data);
        curl_easy_setopt(client->curl, CURLOPT_USERAGENT,
                         "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.15; "
                         "rv:76.0) Gecko/20100101 Firefox/76.0");
        curl_easy_setopt(client->curl, CURLOPT_HTTPHEADER, headers);

        curl_easy_setopt(client->curl, CURLOPT_POST, 1L);
        curl_easy_setopt(client->curl, CURLOPT_POSTFIELDS, "");

        curl_easy_setopt(client->curl, CURLOPT_FOLLOWLOCATION, 1L);

        array(uint8_t) chunk = newArray(uint8_t, 4096);

        curl_easy_setopt(client->curl, CURLOPT_WRITEFUNCTION,
                         write_memory_callback);
        curl_easy_setopt(client->curl, CURLOPT_WRITEDATA, (void *)&chunk);

        CURLcode res = curl_easy_perform(client->curl);

        if (res != CURLE_OK) {
            fprintf(stderr, "curl_easy_perform() failed: %s\n",
                    curl_easy_strerror(res));
        } else {
            fprintf(stderr, "Got data\n%.*s\n", (int)chunk.size, chunk.data);
        }

        // parse json
        cJSON *json = cJSON_ParseWithLength(chunk.data, chunk.size);
        if (json == NULL) {
            const char *error_ptr = cJSON_GetErrorPtr();
            if (error_ptr != NULL) {
                fprintf(stderr,
                        "SLACK CLIENT: Failed to parse /auth.test: %s\n",
                        error_ptr);

                // FIXME: don't panic here
                exit(1);
            }
        }
        {
            const cJSON *ok = cJSON_GetObjectItemCaseSensitive(json, "ok");
            if (cJSON_IsBool(ok) && ok->valueint == 1) {
                fprintf(stderr, "SLACK CLIENT: /auth.test succeeded!\n");

                const cJSON *url =
                    cJSON_GetObjectItemCaseSensitive(json, "url");
                const cJSON *team =
                    cJSON_GetObjectItemCaseSensitive(json, "team");
                const cJSON *user =
                    cJSON_GetObjectItemCaseSensitive(json, "user");
                const cJSON *team_id =
                    cJSON_GetObjectItemCaseSensitive(json, "team_id");
                const cJSON *user_id =
                    cJSON_GetObjectItemCaseSensitive(json, "user_id");

                return (slack_user_info){
                    .url = string_copy_cstring(url->valuestring),
                    .team = string_copy_cstring(team->valuestring),
                    .user = string_copy_cstring(user->valuestring),
                    .team_id = string_copy_cstring(team_id->valuestring),
                    .user_id = string_copy_cstring(user_id->valuestring),
                };
            } else {
                const cJSON *error =
                    cJSON_GetObjectItemCaseSensitive(json, "error");
                if (cJSON_IsString(error)) {
                    fprintf(stderr, "SLACK CLIENT: /auth.test failed: %s\n",
                            error->valuestring);
                } else {
                    fprintf(stderr, "SLACK CLIENT: /auth.test failed: failed "
                                    "to parse 'error' field\n");
                }
            }
        }
        cJSON_Delete(json);

        free(chunk.data);
    }
}

array(slack_channel)
    slack_api_channel_list(slack_client *client, slack_user_info user) {
    if (client->curl) {
        struct curl_slist *headers = NULL;
        headers = curl_slist_append(
            headers, "Content-Type: application/x-www-form-urlencoded");
        headers = curl_slist_append(headers, client->cookie.data);

        curl_easy_setopt(client->curl, CURLOPT_URL,
                         "https://slack.com/api/conversations.list");

        curl_easy_setopt(client->curl, CURLOPT_HTTPAUTH, (long)CURLAUTH_BEARER);
        curl_easy_setopt(client->curl, CURLOPT_XOAUTH2_BEARER,
                         client->token.data);
        curl_easy_setopt(client->curl, CURLOPT_USERAGENT,
                         "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.15; "
                         "rv:76.0) Gecko/20100101 Firefox/76.0");
        curl_easy_setopt(client->curl, CURLOPT_HTTPHEADER, headers);

        curl_easy_setopt(client->curl, CURLOPT_POST, 1L);
        curl_easy_setopt(client->curl, CURLOPT_POSTFIELDS, "limit=10");

        curl_easy_setopt(client->curl, CURLOPT_FOLLOWLOCATION, 1L);

        array(uint8_t) chunk = newArray(uint8_t, 4096);

        curl_easy_setopt(client->curl, CURLOPT_WRITEFUNCTION,
                         write_memory_callback);
        curl_easy_setopt(client->curl, CURLOPT_WRITEDATA, (void *)&chunk);

        CURLcode res = curl_easy_perform(client->curl);

        if (res != CURLE_OK) {
            fprintf(stderr, "curl_easy_perform() failed: %s\n",
                    curl_easy_strerror(res));
        } else {
            fprintf(stderr, "Got data\n%.*s\n", (int)chunk.size, chunk.data);
        }

        // parse json
        cJSON *json = cJSON_ParseWithLength(chunk.data, chunk.size);
        if (json == NULL) {
            const char *error_ptr = cJSON_GetErrorPtr();
            if (error_ptr != NULL) {
                fprintf(
                    stderr,
                    "SLACK CLIENT: Failed to parse /conversations.list: %s\n",
                    error_ptr);

                // FIXME: don't panic here
                exit(1);
            }
        }
        {
            const cJSON *ok = cJSON_GetObjectItemCaseSensitive(json, "ok");
            if (cJSON_IsBool(ok) && ok->valueint == 1) {
                array(slack_channel) channels = newArray(slack_channel, 10);

                fprintf(stderr,
                        "SLACK CLIENT: /conversations.list succeeded!\n");

                const cJSON *json_channels =
                    cJSON_GetObjectItemCaseSensitive(json, "channels");

                const cJSON *json_channel;
                cJSON_ArrayForEach(json_channel, json_channels) {
                    const cJSON *id =
                        cJSON_GetObjectItemCaseSensitive(json_channel, "id");
                    const cJSON *name =
                        cJSON_GetObjectItemCaseSensitive(json_channel, "name");
                    const cJSON *topic =
                        cJSON_GetObjectItemCaseSensitive(json_channel, "topic");
                    const cJSON *purpose = cJSON_GetObjectItemCaseSensitive(
                        json_channel, "purpose");

                    slack_channel channel = {
                        .id = string_copy_cstring(id->valuestring),
                        .name = string_copy_cstring(name->valuestring),
                        .topic = string_copy_cstring(topic->valuestring),
                        .purpose = string_copy_cstring(purpose->valuestring),
                    };

                    pushArray(slack_channel, &channels, channel);
                }

                return channels;
            } else {
                const cJSON *error =
                    cJSON_GetObjectItemCaseSensitive(json, "error");
                if (cJSON_IsString(error)) {
                    fprintf(stderr,
                            "SLACK CLIENT: /conversations.list failed: %s\n",
                            error->valuestring);
                } else {
                    fprintf(stderr,
                            "SLACK CLIENT: /conversations.list failed: failed "
                            "to parse 'error' field\n");
                }
            }
        }
        cJSON_Delete(json);

        free(chunk.data);
    }

    // TODO: create some slack api error type
    return (array(slack_channel)){0};
}

array(slack_message)
    slack_api_message_list(slack_client *client, string channel_id) {

    if (client->curl) {
        struct curl_slist *headers = NULL;
        headers = curl_slist_append(
            headers, "Content-Type: application/x-www-form-urlencoded");
        headers = curl_slist_append(headers, client->cookie.data);

        curl_easy_setopt(client->curl, CURLOPT_URL,
                         "https://slack.com/api/conversations.history");

        curl_easy_setopt(client->curl, CURLOPT_HTTPAUTH, (long)CURLAUTH_BEARER);
        curl_easy_setopt(client->curl, CURLOPT_XOAUTH2_BEARER,
                         client->token.data);
        curl_easy_setopt(client->curl, CURLOPT_USERAGENT,
                         "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.15; "
                         "rv:76.0) Gecko/20100101 Firefox/76.0");
        curl_easy_setopt(client->curl, CURLOPT_HTTPHEADER, headers);

        curl_easy_setopt(client->curl, CURLOPT_POST, 1L);

        char *fields_buffer[2048];
        snprintf(fields_buffer, 2048, "channel=%.*s&limit=10",
                 (int)channel_id.len, channel_id.data);
        fprintf(stderr, "fields_buffer: %s\n", fields_buffer);
        curl_easy_setopt(client->curl, CURLOPT_POSTFIELDS, fields_buffer);

        curl_easy_setopt(client->curl, CURLOPT_FOLLOWLOCATION, 1L);

        array(uint8_t) chunk = newArray(uint8_t, 4096);

        curl_easy_setopt(client->curl, CURLOPT_WRITEFUNCTION,
                         write_memory_callback);
        curl_easy_setopt(client->curl, CURLOPT_WRITEDATA, (void *)&chunk);

        CURLcode res = curl_easy_perform(client->curl);

        if (res != CURLE_OK) {
            fprintf(stderr, "curl_easy_perform() failed: %s\n",
                    curl_easy_strerror(res));
        } else {
            fprintf(stderr, "Got data\n%.*s\n", (int)chunk.size, chunk.data);
        }

        // parse json
        cJSON *json = cJSON_ParseWithLength(chunk.data, chunk.size);
        if (json == NULL) {
            const char *error_ptr = cJSON_GetErrorPtr();
            if (error_ptr != NULL) {
                fprintf(stderr,
                        "SLACK CLIENT: Failed to parse /conversations.history: "
                        "%s\n",
                        error_ptr);

                // FIXME: don't panic here
                exit(1);
            }
        }
        {
            const cJSON *ok = cJSON_GetObjectItemCaseSensitive(json, "ok");
            if (cJSON_IsBool(ok) && ok->valueint == 1) {
                array(slack_message) messages = newArray(slack_message, 10);

                fprintf(stderr,
                        "SLACK CLIENT: /conversations.history succeeded!\n");

                const cJSON *json_messages =
                    cJSON_GetObjectItemCaseSensitive(json, "messages");

                const cJSON *json_message;
                cJSON_ArrayForEach(json_message, json_messages) {
                    const cJSON *type =
                        cJSON_GetObjectItemCaseSensitive(json_message, "type");
                    const cJSON *user =
                        cJSON_GetObjectItemCaseSensitive(json_message, "user");
                    const cJSON *text =
                        cJSON_GetObjectItemCaseSensitive(json_message, "text");

                    if (strcmp(type->valuestring, "message") == 0) {
                        slack_message message = {
                            .type = SLACK_MESSAGE_TYPE_MESSAGE,
                            .normal =
                                (slack_message_normal){
                                    .user =
                                        string_copy_cstring(user->valuestring),
                                    .text =
                                        string_copy_cstring(text->valuestring),
                                },
                        };

                        pushArray(slack_message, &messages, message);
                    } else {
                        fprintf(stderr,
                                "SLACK CLIENT: unknown message type: %s\n",
                                type->valuestring);
                    }
                }

                return messages;
            } else {
                const cJSON *error =
                    cJSON_GetObjectItemCaseSensitive(json, "error");
                if (cJSON_IsString(error)) {
                    fprintf(stderr,
                            "SLACK CLIENT: /conversations.history failed: %s\n",
                            error->valuestring);
                } else {
                    fprintf(
                        stderr,
                        "SLACK CLIENT: /conversations.history failed: failed "
                        "to parse 'error' field\n");
                }
            }
        }
        cJSON_Delete(json);

        free(chunk.data);
    }

    // TODO: create some slack api error type
    return (array(slack_message)){0};
}

array(slack_user) slack_api_user_list(slack_client *client, string team_id) {
    if (client->curl) {
        struct curl_slist *headers = NULL;
        headers = curl_slist_append(
            headers, "Content-Type: application/x-www-form-urlencoded");
        headers = curl_slist_append(headers, client->cookie.data);

        curl_easy_setopt(client->curl, CURLOPT_URL,
                         "https://slack.com/api/users.list");

        curl_easy_setopt(client->curl, CURLOPT_HTTPAUTH, (long)CURLAUTH_BEARER);
        curl_easy_setopt(client->curl, CURLOPT_XOAUTH2_BEARER,
                         client->token.data);
        curl_easy_setopt(client->curl, CURLOPT_USERAGENT,
                         "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.15; "
                         "rv:76.0) Gecko/20100101 Firefox/76.0");
        curl_easy_setopt(client->curl, CURLOPT_HTTPHEADER, headers);

        curl_easy_setopt(client->curl, CURLOPT_POST, 1L);

        char *fields_buffer[2048];
        snprintf(fields_buffer, 2048, "team_id=%.*s", (int)team_id.len,
                 team_id.data);
        fprintf(stderr, "fields_buffer: %s\n", fields_buffer);
        curl_easy_setopt(client->curl, CURLOPT_POSTFIELDS, fields_buffer);

        curl_easy_setopt(client->curl, CURLOPT_FOLLOWLOCATION, 1L);

        // TODO: don't allocate this on every request
        array(uint8_t) chunk = newArray(uint8_t, 4096);

        curl_easy_setopt(client->curl, CURLOPT_WRITEFUNCTION,
                         write_memory_callback);
        curl_easy_setopt(client->curl, CURLOPT_WRITEDATA, (void *)&chunk);

        CURLcode res = curl_easy_perform(client->curl);

        if (res != CURLE_OK) {
            fprintf(stderr, "curl_easy_perform() failed: %s\n",
                    curl_easy_strerror(res));
        } else {
            fprintf(stderr, "Got data\n%.*s\n", (int)chunk.size, chunk.data);
        }

        // parse json
        cJSON *json = cJSON_ParseWithLength(chunk.data, chunk.size);
        if (json == NULL) {
            const char *error_ptr = cJSON_GetErrorPtr();
            if (error_ptr != NULL) {
                fprintf(stderr,
                        "SLACK CLIENT: Failed to parse /users.list: %s\n",
                        error_ptr);

                // FIXME: don't panic here
                exit(1);
            }
        }
        {
            const cJSON *ok = cJSON_GetObjectItemCaseSensitive(json, "ok");
            if (cJSON_IsBool(ok) && ok->valueint == 1) {
                array(slack_user) users = newArray(slack_user, 10);

                fprintf(stderr, "SLACK CLIENT: /users.list succeeded!\n");

                const cJSON *json_users =
                    cJSON_GetObjectItemCaseSensitive(json, "members");

                const cJSON *json_user;
                cJSON_ArrayForEach(json_user, json_users) {
                    const cJSON *user_id =
                        cJSON_GetObjectItemCaseSensitive(json_user, "id");
                    const cJSON *name =
                        cJSON_GetObjectItemCaseSensitive(json_user, "name");
                    const cJSON *real_name = cJSON_GetObjectItemCaseSensitive(
                        json_user, "real_name");

                    slack_user user = {
                        .id = string_copy_cstring(user_id->valuestring),
                        .name = string_copy_cstring(name->valuestring),
                        .real_name =
                            string_copy_cstring(real_name->valuestring),
                    };

                    pushArray(slack_user, &users, user);
                }

                return users;
            } else {
                const cJSON *error =
                    cJSON_GetObjectItemCaseSensitive(json, "error");
                if (cJSON_IsString(error)) {
                    fprintf(stderr, "SLACK CLIENT: /users.list failed: %s\n",
                            error->valuestring);
                } else {
                    fprintf(stderr, "SLACK CLIENT: /users.list failed: failed "
                                    "to parse 'error' field\n");
                }
            }
        }
        cJSON_Delete(json);

        free(chunk.data);
    }

    // TODO: create some slack api error type
    return (array(slack_user)){0};
}

slack_client slack_init_client(string token, string cookie) {
    CURL *curl = curl_easy_init();

    slack_client client =
        (slack_client){.curl = curl, .token = token, .cookie = cookie};

    return client;
}

#endif
#endif
